import private/nix

import nimblepkg/common,
  nimblepkg/download,
  nimblepkg/packageinfo,
  nimblepkg/options,
  nimblepkg/version,
  nimblepkg/packageparser,
  nimblepkg/cli

import std/[algorithm, deques, httpclient, json, monotimes, os, osproc, parseutils, random, sequtils, streams, strutils, tables, times, uri]

const githubPackagesUrl =
  "https://raw.githubusercontent.com/nim-lang/packages/master/packages.json"

proc registryCachePath: string =
  result = getEnv("XDG_CACHE_HOME")
  if result == "":
    let home = getEnv("HOME")
    if home == "":
      result = getEnv("TMPDIR", "/tmp")
    else:
      result = home / ".cache"
  result.add "/packages.json"

func isGitUrl(uri: Uri): bool =
  uri.path.endsWith(".git") or
    uri.scheme == "git" or
    uri.scheme.startsWith("git+")

proc gitLsRemote(url: string): seq[tuple[tag: string, rev: string]] =
  var lines = execProcess(
      "git",
      args = ["ls-remote", "--tags", url],
      options = {poUsePath},
    )
  result.setLen(lines.countLines.pred)
  var off = 0
  for i in 0..result.high:
    off.inc parseUntil(lines, result[i].rev, {'\x09'}, off)
    off.inc skipWhiteSpace(lines, off)
    off.inc skipUntil(lines, '/', off).succ
    off.inc skipUntil(lines, '/', off).succ
    off.inc parseUntil(lines, result[i].tag, {'\x0a'}, off).succ

proc matchRev(url: string; wanted: VersionRange): tuple[tag: string, rev: string] =
  let pairs = gitLsRemote(url)
  var resultVersion: Version
  for (tag, rev) in pairs:
    var tagVer = Version(tag)
    if tagVer.withinRange(wanted) and resultVersion < tagVer:
      resultVersion = tagVer
      result = (tag, rev)
  if result.rev == "" and pairs.len > 0:
    result = pairs[pairs.high]

proc collectMetadata(data: JsonNode; options: Options) =
  let storePath = data["path"].getStr
  var packageNames = newJArray()
  for (kind, path) in walkDir(storePath):
    if kind in {pcFile, pcLinkToFile} and path.endsWith(".nimble"):
      var (dir, name, ext) = splitFile(path)
      packageNames.add %name
  if packageNames.len == 0:
    quit("no .nimble files found in " & storePath)
  data["packages"] = packageNames
  var
    nimbleFilePath = findNimbleFile(storePath, true)
    pkg = readPackageInfo(nimbleFilePath, options)
  data["srcDir"] = %pkg.srcDir

proc prefetchGit(uri: Uri; version: VersionRange; options: Options): JsonNode =
  var
    uri = uri
    subdir = ""
  uri.scheme.removePrefix("git+")
  if uri.query != "":
    if uri.query.startsWith("subdir="):
      subdir = uri.query[7 .. ^1]
    uri.query = ""
  let url = $uri
  let (tag, rev) = matchRev(url, version)
  var args = @["--quiet", "--fetch-submodules", "--url", url]
  if rev != "":
    args.add "--rev"
    args.add rev
  let dump = execProcess(
    "nix-prefetch-git",
    args = args,
    options = {poUsePath})
  try: result = parseJson dump
  except JsonParsingError:
    stderr.writeLine "failed to parse output of nix-prefetch-git ", join(args, " ")
    quit(dump)
  if subdir != "":
    result["subdir"] = %* subdir
  result["method"] = %"git"
  if tag != "":
    result["ref"] = %tag
  collectMetadata(result, options)

proc containsPackageUri(lockAttrs: JsonNode; pkgUri: string): bool =
  for e in lockAttrs.items:
    if e["url"].getStr == pkgUri:
      return true

proc containsPackage(lockAttrs: JsonNode; pkgName: string): bool =
  for e in lockAttrs.items:
    for other in e["packages"].items:
      if pkgName == other.getStr:
        return true

proc collectRequires(pending: var Deque[PkgTuple]; options: Options; pkgPath: string) =
  var
    nimbleFilePath = findNimbleFile(pkgPath, true)
    pkg = readPackageInfo(nimbleFilePath, options)
  for pair in pkg.requires:
    if pair.name != "nim" and pair.name != "compiler":
      pending.addLast(pair)

var globalRegistry: JsonNode

proc getPackgeUri(name: string): tuple[uri: string, meth: string] =
  if globalRegistry.isNil:
    let registryPath = registryCachePath()
    if fileExists(registryPath):
      globalRegistry = parseFile(registryPath)
    else:
      let client = newHttpClient()
      var raw = client.getContent(githubPackagesUrl)
      close(client)
      writeFile(registryPath, raw)
      globalRegistry = parseJson(raw)
  var
    name = name
    i = 0
  while i < globalRegistry.len:
    var e = globalRegistry[i]
    if e["name"].getStr == name:
      if e.hasKey "alias":
        var alias = e["alias"].getStr
        doAssert alias != name
        name = alias
        i = 0
      else:
        try:
          return (e["url"].getStr, e["method"].getStr,)
        except CatchableError:
          quit("Failed to parse shit JSON " & $e)
    inc i

proc generateLockfile(options: Options): JsonNode =
  result = newJObject()
  var
    deps = newJArray()
    pending: Deque[PkgTuple]
  collectRequires(pending, options, getCurrentDir())
  while pending.len > 0:
    let batchLen = pending.len
    for i in 1..batchLen:
      var pkgData: JsonNode
      let pkg = pending.popFirst()
      if pkg.name == "nim" or pkg.name == "compiler":
        continue
      var uri = parseUri(pkg.name)
      if uri.scheme == "":
        if not deps.containsPackage(pkg.name):
          pending.addLast(pkg)
      elif not deps.containsPackageUri(pkg.name):
        if uri.isGitUrl:
          pkgData = prefetchGit(uri, pkg.ver, options)
        else:
          quit("unhandled URI " & $uri)
        collectRequires(pending, options, pkgData["path"].getStr)
        deps.add pkgData

    if batchLen == pending.len:
      var
        pkg = pending.popFirst()
        info = getPackgeUri(pkg.name)
      case info.meth
      of "git":
        stderr.writeLine "prefetch ", info.uri
        var pkgData = prefetchGit(parseUri info.uri, pkg.ver, options)
        collectRequires(pending, options, pkgData["path"].getStr)
        deps.add pkgData
      else:
        quit("unhandled fetch method " & $info.meth & " for " & info.uri)
  sort(deps.elems)
  result["depends"] = deps

proc main =
  var options = parseCmdLine()
    # parse nimble options, not recommended
  if options.action.typ != actionCustom:
    options.action = Action(typ: actionCustom)
  var lockInfo = generateLockfile(options)
  stdout.writeLine lockInfo

main()
